/* tty line ISDN status monitor
 *
 * (c) 1995-97 Volker Götz
 *
 * $Id: imontty.h,v 1.2 2000/06/29 17:38:26 akool Exp $
 */

#define		IM_BUFSIZE	10 + (22 * ISDN_MAX_CHANNELS)
#define		IM_VERSION	"0.5"
